==> SSOAP_0.4-3.tar.gz: SSOAP/R/wsdlTypes.R <==
processWSDLTypes =
   # Process all the elements in types (hopefully <schema> elements)
   # and merge. We may want to keep these as a list indexed by namespace/schema identifier.
   # Is this still true?
  
   # Originally, we just processed the first one.
   # types = processWSDLTypes(doc[["types"]][[1]], doc)


  # We are called with the doc[["types"]] argument
  # That should contain a schema.  The schema
  # can have elements within it (e.g. interop.wsdl) or
  # import statements which give other schema, e.g. eutils.wsdl
  # or both import statements and new type defintions (e.g. KEGG.wsdl).
  #

  # This can be called recursively for sub-schema, i.e. schema defined within the
  # <types><schema>...</schema></types>

  # This is (initially) called with the top-level <types> which should contain one
  # <schema> node.

function(node, doc, nsDefs = xmlNamespaceDefinitions(doc))
{
  if(is.null(node))
      return(list())

     # Loop over the children of this node and process each element.
 ans <- list()
 lapply(1:xmlSize(node),
                function(i) {
                   el = node[[i]]

                   if(xmlName(el) == "annotation")
                     return(NULL)
                   
                   if(xmlName(el) %in% "schema") {
                      ns = xmlGetAttr(el, "targetNamespace")
                      # cat("processing (sub) schema", xmlGetAttr(el, "targetNamespace"), "\n")
                      o = processWSDLTypes(el, doc)
                      o = new("SchemaTypes", o)
                      if(length(o))
                        ans[[ns]] <<-  o
                   } else {
                      n = xmlGetAttr(el, "name")
                      o = processWSDLType(el)
                      ans[[n]] <<- o
                    }

                   o
                 })

         # Fix the names on these types to avoid the schema.
 #             types = unlist(types, recursive = FALSE)
 #names(types) = sapply(types, function(x) x$name)
 # types

  if(any(sapply(ans, inherits, "SchemaTypes")))
    ans = new("SchemaCollection", ans)
  
  ans
}




# This material relates to XML schema and processing of types defined within
# schema. Ideally, we will separate this into a separate package and use the
# results in different ways, not just for SOAP. And we will provide a more
# extensive framework for dealing with all of the schema details rather than
# the rather limited but common ones supported here.
#
#  See the O'Reilly book XML Schemas by Eric van der Vlist for more information
#  on schemas, or look on the Web (e.g. www.xml.com/schemas)
#

# Failed  complexType & attribute.
#   Assay_molTypeType
#   Assembly_currentType


processWSDLType =
  #
  # This is intended to create an R description for
  # XML Schema nodes that actuall define a type.
  #
  #
  # This currently deals with a very small subset of XML schema
  # specifications.  These are the common ones, but more work 
  # is definitely needed.
  #

  #  simpleType
  #  complexType
  #  complexContent

   # sequence

  # Doesn't look like we use types.
function(type, types, namespaceDefs = list())
{
# Rationalize this 

  name = xmlGetAttr(type, "name", "")

  if(name == "simpleType") {
    if(xmlSize(type) > 0 && xmlName(type[[1]]) == "restriction") {
         # Assume enumeration values for present.
       def <- new("EnumValuesDef", xmlSApply(type[[1]],  xmlGetAttr, "value"))
    } else
      def <- "xsd:string"
    return(def)
  }

  if(xmlName(type) %in% c("complexContent", "element"))
    tmp = type
  else if(xmlName(type) == "complexType" && xmlName(type[[1]]) == "all")
    tmp = type
  else
    tmp <- type[["complexContent"]]

  if(xmlName(type) == "element") {
       return(processSchemaElement(tmp))
  }  else if(!is.null(tmp) && !is.null(tmp[["all"]])) {

             # a struct-like definition with slots.
     tmp = tmp[["all"]]

     slotTypes = xmlApply(tmp, function(x) {
                                  tt = xmlGetAttr(x, "type")
                                   getSOAPType(tt, types) #XXX , namespaceDefs)
                               })
     names(slotTypes) = xmlSApply(tmp, function(x) xmlGetAttr(x, "name"))
     def = ClassDef(name, slotTypes)

  } else if(!is.null(tmp) && !is.null(tmp[["extension"]])) {

     base = xmlGetAttr(tmp[["extension"]], "base")
                      # Now get the extensions.
     def = ClassDef(name, list(), new("ExtendedClassDefinition"))
     def@base = base
     tmp = tmp[[1]]
     if(xmlSize(tmp) > 0 && xmlName(tmp[[1]]) == "sequence") {
                      #XXXXX Taken from sequence below. Merge into 1.
      slotTypes = xmlApply(tmp[[1]], function(x) SOAPType(xmlGetAttr(x, "type"), namespaceDefs = namespaceDefs))
      names(slotTypes) = xmlSApply(tmp[[1]], function(x) xmlGetAttr(x, "name"))
      def@slotTypes = slotTypes       
     }

  } else if(!is.null(tmp)) {
    
                   # Currently we interpret this as an Array.
                   # So get the <restriction><attribute> element
                   # and take the arrayType and ref attribution from that.
    restriction = tmp[["restriction"]]
    a = restriction[["attribute"]]
    def = ArrayType(xmlGetAttr(a, "arrayType"), namespaceDefs = namespaceDefs)
    def@name = name
    
  } else if(!is.null(tmp <- type[["all"]])) {
    
                    # a struct-like definition with slots.
     slotTypes = xmlApply(tmp, function(x) xmlGetAttr(x, "type"))
     names(slotTypes) = xmlSApply(tmp, function(x) xmlGetAttr(x, "name"))
     def = ClassDef(name, slotTypes)
     
  } else if(!is.null(tmp <- type[["sequence"]])) {
    
                    # <complexType><sequence>
                    # Connect this with the code for the "all" case.
                    # Need to resolve the type if it is not a primitive.
                    # XXX also want the minOccurs and maxOccurs
      slotTypes = xmlApply(tmp, function(x) SOAPType(xmlGetAttr(x, "type"), namespaceDefs = namespaceDefs))
      names(slotTypes) = xmlSApply(tmp, function(x) xmlGetAttr(x, "name"))           
      def = ClassDef(name, slotTypes)
      
  } else if(!is.null(tmp <- type[["choice"]])) {
    
      slotTypes = xmlApply(tmp, function(x) xmlGetAttr(x, "type"))
      names(slotTypes) = xmlSApply(tmp, function(x) xmlGetAttr(x, "name"))               
      if(xmlSize(tmp) == 1)
        def = ClassDef(name, slotTypes[1])
      else          #XXX Define UnionDef
        def = UnionDef(name, slotTypes)
      
  } else if(!is.null(tmp <- type[["simpleContent"]])) {
      if(xmlName(tmp[[1]]) == "extension") {
        
                      # Assume the children are attributes for the moment.
         attrTypes = xmlApply(tmp[[1]], processWSDLType)
         names(attrTypes) = xmlSApply(tmp[[1]], function(x) xmlGetAttr(x, "name"))
                      #XXX Add base.
         base = xmlGetAttr(tmp[[1]], "base")         
         def <- ClassDef(name, attrTypes)
         
      } else
          def <- NULL
  } else {
     warning("Failed to handle node ", name, " of type ", xmlName(type), if(xmlSize(type) > 0) c("&", xmlName(type[[1]])), " in processWSDLType")
     return(NULL)
      #      def <- NULL
  }
  
  structure(list(name = name, definition = def),
            class = "WSDLTypeDescription")
}


# Children
# no children - simple element  esearch.xsd  Count.
# complexType with a sequence  egquery.xsd  Result
# simpletype

setClass("SchemaCollection", contains = "list")
setClass("SchemaTypes", contains = "list")

setClass("SchemaElement", representation(name = "character"))
setClass("SimpleElement", representation(type = "character"), contains = "SchemaElement")
setClass("Element", representation(elements = "character",
                                   attributes = "list"), contains = "SchemaElement")

processSchemaElement =
function(element, name = xmlGetAttr(element, "name"), namespaceDefs = list())
{
  return(element)
  
  if(xmlSize(element) == 0) {
    obj =  new("SimpleElement", name = name)
  } else {
     # complexType
    if(xmlName(element[[1]]) == "complexType")
        element
  }

  obj
}

==> SSOAP_0.4-6.tar.gz: SSOAP/R/wsdlTypes.R <==
processWSDLTypes =
   # Process all the elements in types (hopefully <schema> elements)
   # and merge. We may want to keep these as a list indexed by namespace/schema identifier.
   # Is this still true?
  
   # Originally, we just processed the first one.
   # types = processWSDLTypes(doc[["types"]][[1]], doc)


  # We are called with the doc[["types"]] argument
  # That should contain a schema.  The schema
  # can have elements within it (e.g. interop.wsdl) or
  # import statements which give other schema, e.g. eutils.wsdl
  # or both import statements and new type defintions (e.g. KEGG.wsdl).
  #

  # This can be called recursively for sub-schema, i.e. schema defined within the
  # <types><schema>...</schema></types>

  # This is (initially) called with the top-level <types> which should contain one
  # <schema> node.

function(node, doc, nsDefs = xmlNamespaceDefinitions(doc))
{
  if(is.null(node))
      return(list())

     # Loop over the children of this node and process each element.
 ans <- list()
 lapply(1:xmlSize(node),
                function(i) {
                   el = node[[i]]

                   if(xmlName(el) == "annotation")
                     return(NULL)
                   
                   if(xmlName(el) %in% "schema") {
                      ns = xmlGetAttr(el, "targetNamespace")
                      # cat("processing (sub) schema", xmlGetAttr(el, "targetNamespace"), "\n")
                      o = processWSDLTypes(el, doc)
                      o = new("SchemaTypes", o)
                      if(length(o))
                        ans[[ns]] <<-  o
                   } else {
                      n = xmlGetAttr(el, "name")
                      o = processWSDLType(el)
                      ans[[n]] <<- o
                    }

                   o
                 })

         # Fix the names on these types to avoid the schema.
 #             types = unlist(types, recursive = FALSE)
 #names(types) = sapply(types, function(x) x$name)
 # types

  if(any(sapply(ans, inherits, "SchemaTypes")))
    ans = new("SchemaCollection", ans)
  
  ans
}




# This material relates to XML schema and processing of types defined within
# schema. Ideally, we will separate this into a separate package and use the
# results in different ways, not just for SOAP. And we will provide a more
# extensive framework for dealing with all of the schema details rather than
# the rather limited but common ones supported here.
#
#  See the O'Reilly book XML Schemas by Eric van der Vlist for more information
#  on schemas, or look on the Web (e.g. www.xml.com/schemas)
#

# Failed  complexType & attribute.
#   Assay_molTypeType
#   Assembly_currentType


processWSDLType =
  #
  # This is intended to create an R description for
  # XML Schema nodes that actually define a type.
  #
  #
  # This currently deals with a very small subset of XML schema
  # specifications.  These are the common ones, but more work 
  # is definitely needed.
  #

  #  simpleType
  #  complexType
  #  complexContent

   # sequence

  # Doesn't look like we use types.

  # Fix for
  #   simpleType with "attribute" child.
  #   simpleType with restriction.
function(type, types, namespaceDefs = list())
{
# Rationalize this 

  name = xmlGetAttr(type, "name", "")

  if(xmlSize(type) > 0 && "annotation" == xmlName(type[[1]])) # drop the annotation sub-node.
    xmlChildren(type) = xmlChildren(type)[-1]

  
  if(name == "simpleType" || xmlName(type) == "simpleType") {
    if(xmlSize(type) > 0 && xmlName(type[[1]]) == "restriction") {
       if(xmlSize(type[[1]]) > 0) {
           # Assume enumeration values for present.
         def <- new("EnumValuesDef", name = name, values = xmlSApply(type[[1]],  xmlGetAttr, "value"))
       } else {
         # e.g. from eBaySvc.wsdl
         # <xs:simpleType name="DisputeIDType">
         #   <xs:restriction base="xs:string"/>
         # </xs:simpleType>
         def = new("ExtendedClassDefinition", name = name, base = xmlGetAttr(type[[1]], "base"))
       }
   } else
       def <- "xsd:string"

    return(def)
  }

  if(xmlName(type) %in% c("complexContent", "element"))
    tmp = type
  else if(xmlName(type) == "complexType" && xmlName(type[[1]]) == "all")
    tmp = type
  else if(xmlName(type) == "attribute" && xmlSize(type) == 0) {
     return(new("AttributeDef", name = name, type = xmlGetAttr(type, "type"), use = xmlGetAttr(type, "use", "optional")))
  }
  else
    tmp <- type[["complexContent"]]

  if(xmlName(type) == "element") {
       return(processSchemaElement(tmp))
  } else if(!is.null(tmp) && !is.null(tmp[["all"]])) {
        # Connect this with the other case !is.null(tmp <- type[["all"]]) below.
    
             # a struct-like definition with slots.
     tmp = tmp[["all"]]

     slotTypes = xmlApply(tmp, function(x) {
                                   # just return the SOAP type. We'll resolve it later.
                                  tt = xmlGetAttr(x, "type")
                                  SOAPType(tt, nsuri = xmlNamespace(x), namespaceDefs = namespaceDefs)
                                    #getArrayElementTypeFromName
                               })
     names(slotTypes) = xmlSApply(tmp, function(x) xmlGetAttr(x, "name"))
     def = ClassDef(name, slotTypes)

  } else if(!is.null(tmp) && !is.null(tmp[["extension"]])) {

     base = xmlGetAttr(tmp[["extension"]], "base")
                      # Now get the extensions.
     def = ClassDef(name, list(), new("ExtendedClassDefinition"))
     def@base = base
     tmp = tmp[[1]]
     if(xmlSize(tmp) > 0 && xmlName(tmp[[1]]) == "sequence") {
                      #XXXXX Taken from sequence below. Merge into 1.
      slotTypes = xmlApply(tmp[[1]], function(x) SOAPType(xmlGetAttr(x, "type"), nsuri = xmlNamespace(x), namespaceDefs = namespaceDefs))
      names(slotTypes) = xmlSApply(tmp[[1]], xmlGetAttr, "name")
      def@slotTypes = slotTypes       
     }

  } else if(!is.null(tmp)) {
    
                   # Currently we interpret this as an Array.
                   # So get the <restriction><attribute> element
                   # and take the arrayType and ref attribution from that.
            # In interop.wsdl, we also have a <sequence> <element> ... node within this restriction.
    restriction = tmp[["restriction"]]
    a = restriction[["attribute"]]
    def = ArrayType(xmlGetAttr(a, "arrayType"), ns = xmlNamespace(a), namespaceDefs = namespaceDefs)
    def@name = name
    
  } else if(!is.null(tmp <- type[["all"]])) {
    
                    # a struct-like definition with slots.
     slotTypes = xmlApply(tmp, xmlGetAttr, "type")
     names(slotTypes) = xmlSApply(tmp, xmlGetAttr, "name")
     def = ClassDef(name, slotTypes)
     
  } else if(!is.null(tmp <- type[["sequence"]])) {
    
                    # <complexType><sequence>
                    # Connect this with the code for the "all" case.
                    # Need to resolve the type if it is not a primitive.
                    # XXX also want the minOccurs and maxOccurs
      slotTypes = xmlApply(tmp, function(x) SOAPType(xmlGetAttr(x, "type"), nsuri = xmlNamespace(x), namespaceDefs = namespaceDefs))
      names(slotTypes) = ids = xmlSApply(tmp, function(x) xmlGetAttr(x, "name", NA))
      slotTypes = slotTypes[ !is.na(ids) ] 
      def = ClassDef(name, slotTypes)
      
  } else if(!is.null(tmp <- type[["choice"]])) {
    
      slotTypes = xmlApply(tmp, xmlGetAttr, "type")
      names(slotTypes) = xmlSApply(tmp,  xmlGetAttr, "name")
      if(xmlSize(tmp) == 1)
        def = ClassDef(name, slotTypes[1])
      else          #XXX Define UnionDef
        def = UnionDef(name, slotTypes)
      
  } else if(!is.null(tmp <- type[["simpleContent"]])) {
      if(xmlName(tmp[[1]]) == "extension") {
        
                      # Assume the children are attributes for the moment.
         attrTypes = xmlApply(tmp[[1]], processWSDLType)
         names(attrTypes) = xmlSApply(tmp[[1]],  xmlGetAttr, "name")
                      #XXX Add base.
         base = xmlGetAttr(tmp[[1]], "base")         
         def <- ClassDef(name, attrTypes)
         
      } else
          def <- NULL
  } else {
     warning("Failed to handle node ", name, " of type ", xmlName(type), if(xmlSize(type) > 0) c(" & ", xmlName(type[[1]])), " in processWSDLType")
     return(NULL)
      #      def <- NULL
  }
 
  structure(list(name = name, definition = def),
            class = "WSDLTypeDescription")
}


# Children
# no children - simple element  esearch.xsd  Count.
# complexType with a sequence  egquery.xsd  Result
# simpletype

setClass("SchemaCollection", contains = "list")
setClass("SchemaTypes", contains = "list")

setClass("SchemaElement", representation(name = "character"))
setClass("SimpleElement", representation(type = "character"), contains = "SchemaElement")
setClass("Element", representation(elements = "character",
                                   attributes = "list"), contains = "SchemaElement")


setClass("AttributeDef", representation(name = "character", type = "character",
                                        use = "character"  # prohibited, optional, required
                                        ),
                         prototype = list(use = "optional"))

processSchemaElement =
function(element, name = xmlGetAttr(element, "name"), namespaceDefs = list())
{
  return(element)
  
  if(xmlSize(element) == 0) {
    obj =  new("SimpleElement", name = name)
  } else {
     # complexType
    if(xmlName(element[[1]]) == "complexType")
        element
  }

  obj
}
